// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <cmath>

#include <saotrace/io/fits/fits.h>
#include <randlib_c/randlib.h>

#include  "EsaLinPowSymDB.h"

namespace FITS = SAOTrace::IO::FITS;

namespace Scatter {

    bool
    EsaLinPowSymDB::db_matches( const std::string& filename ) {

	FITS::TableFH table( filename );

	return table.has_column( "ESA" )
	    && table.has_column( "PDIST" )
	    && table.has_column( "P0" )
	    && table.has_column( "PTOP" )
	    && table.has_column( "PDELTA" )
	    && table.has_column( "POWTOP" )
	    && table.has_column( "NORMTOP" )
	    ;

    }



    EsaLinPowSymDB::EsaLinPowSymDB( const std::string& filename )
	: ScatterDB( filename ) {

	std::vector <FITS::ColumnBase*> columns;

	FITS::Column<double> esa     ( "ESA"		, _esa	      , columns );
	FITS::Column<double> pdist   ( "PDIST"	, _pdist.data , columns );
	FITS::Column<double> p0      ( "P0"		, _p0 	      , columns );
	FITS::Column<double> ptop    ( "PTOP"		, _ptop       , columns );
	FITS::Column<double> pdelta  ( "PDELTA"	, _pdelta     , columns );
	FITS::Column<double> powtop  ( "POWTOP"	, _powtop     , columns );
	FITS::Column<double> normtop ( "NORMTOP"	, _normtop    , columns );

	_nrows = FITS::read_table_columns( _file, columns );
	_pdist.ncols = pdist.extent[0];
    }


    bool
    EsaLinPowSymDB::lookUpScatterAngle( double energy, double sin_alpha,
					double rannum, double& theta ) {

	double esa = energy * sin_alpha;

	std::pair< ColData::iterator, ColData::iterator > range;
	std::pair< long, long > idx;

	bool is_extrapolated = find_bounds( _esa, esa, range, idx );

	rannum = 2.0 * ( rannum - 0.5 );
	bool is_neg = ( rannum < 0.0 );

	if ( is_neg )
	    rannum = -1.0 * rannum;

	double theta_esa_lo = lookUpTheta( idx.first, rannum );
	double theta_esa_hi = lookUpTheta( idx.second, rannum );


	if ( theta_esa_lo < 0.0 ) theta_esa_lo = 0.0;
	if ( theta_esa_hi < 0.0 ) theta_esa_hi = 0.0;

	theta = interpolate( esa, *range.first, theta_esa_lo, *range.second, theta_esa_hi );

	if ( is_neg )
	    theta *= -1;

	return is_extrapolated;
    }

    double
    EsaLinPowSymDB::lookUpTheta( const long idx, double rannum ) {

	if ( rannum <= _p0[idx] ) {

	    return 0.0;

	} else if ( rannum <= _ptop[idx] ) {

	    rannum = ( rannum - _p0[idx] ) / _pdelta[idx];

	    long jdx = (long) rannum;

	    long ndist = _pdist.ncols;

	    if ( jdx >= ndist )
		jdx = ndist - 1;

	    rannum = rannum - (double) jdx;
	    double t = _pdist( idx, jdx );

	    return rannum * _pdist( idx, jdx + 1) + ( 1.0 - rannum ) * t ;

	} else {

	    double rtop = 0.99999; // Ouch!  this was in the old code...

	    if ( rannum < rtop ) {

		return pow( _normtop[idx] * ( 1.0 - rannum ), _powtop[idx] );

	    } else {

		return pow( _normtop[idx] * ( 1.0 - rtop ), _powtop[idx] );

	    }

	}

    }

}
